package internal

import (
	"github.com/icinga/icinga-go-library/version"
)

// Version contains version and Git commit information.
//
// The placeholders are replaced on `git archive` using the `export-subst` attribute.
var Version = version.Version("1.3.0", "v1.3.0", "a3bfecbe94783ecf9a585ee8260740fe2e12d68b")
